# -*- encoding: utf-8 -*-
# @Time    :   2025/10/12 12:48:14
# @File    :   utils.py
# @Author  :   ciaoyizhen
# @Contact :   yizhen.ciao@gmail.com
# @Function:   通用工具
import json
from functools import wraps
from threading import Lock

def singleton(cls):
    """线程安全的单例模式装饰器"""
    instances = {}
    lock = Lock()

    @wraps(cls)
    def get_instance(*args, **kwargs):
        # 双重检查锁，防止多线程竞争
        if cls not in instances:
            with lock:
                if cls not in instances:
                    instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return get_instance


def return_to_jsonl(file_path, encoding="utf-8", ensure_ascii=False):
    def decorator(func):
        
        @wraps(func)
        def wrapper(*arg, **kwargs):
            single_result = func(*arg, **kwargs)
            error_msg = f"被装饰器的函数需要有返回，并且必须是str或dict"
            if (single_result is None):
                raise RuntimeError(error_msg)
            
            if isinstance(single_result, dict):
                single_result = json.dumps(single_result, ensure_ascii=ensure_ascii)
            elif isinstance(single_result, str):
                pass
            else:
                raise RuntimeError(error_msg)
            
            with open(file_path, "a", encoding=encoding) as f:
                f.write(single_result + "\n")
            
        return wrapper
    
    return decorator


# -*- encoding: utf-8 -*-
# @Time    :   2025/10/11 19:57:41
# @File    :   utils.py
# @Author  :   ciaoyizhen
# @Contact :   yizhen.ciao@gmail.com
# @Function:   读取文件和保存文件
import os
import json
from pathlib import Path
from typing import List, Dict, Literal
from tqdm import tqdm
from ..utils.log_manage import base_logger

    
def read_file(file_name: str|Path, *, output_type="list", file_type=None, main_key_column=None, encoding="utf-8", disable_tqdm=False, **kwargs)-> List|Dict:
    """读取文件，根据传参来判断读取的方式
    最终返回完整的一个list

    Args:
        file_name (str|Path): 文件路径
        
        output_type (Literal["list", "dict", "set"]): 返回类型,当该值为dict的时候,需要指定output_type
        file_type (str): 文件类型,请使用`json`,`jsonl`,`xlsx`,`csv`,`txt`
        encoding (str): 文件编码方式
        key_columns (list): 需要取的列名
        main_key_column (str): 当返回为dict时,这个为key,value为其他的值,类型为dict
        output_type (Literal["list", "dict"]): 返回类型,当该值为dict的时候
        disable_tqdm (bool): 是否关闭进度条
        
        kwargs: 其他参数
            - sheet_name (str): 读取xlsx时，可以指定读取哪个sheet_name
        
    Returns:
        list|dict|set: 根据output_type返回List|Dict|set
    """
    
    if isinstance(file_name, str):
        file_name = Path(file_name)
        
    if file_type is None:
        file_type = file_name.suffix.lstrip(".")

    match output_type:
        case "list":
            return_data = []
        case "dict":
            return_data = {}
        case "set":
            return_data = set()
        case _:
            raise RuntimeError(f"output_type 传入了一个不可预知的参数:{output_type=}\n目前仅允许`list`, `dict`, `set`")
        
    
        
    match file_type:
        case "jsonl":
            with file_name.open("r", encoding=encoding) as f:
                for line in tqdm(f.readlines(), disable=disable_tqdm):
                    if line := line.strip():
                        line = json.loads(line)
                        if isinstance(return_data, list):
                            return_data.append(line)
                        elif isinstance(return_data, dict):
                            if main_key_column not in line:
                                raise RuntimeError(f"对象没有{main_key_column=}\n原始数据:{row}")

                            value = line[main_key_column]
                            return_data[value] = line
                        elif isinstance(return_data, set):
                            if main_key_column not in line:
                                raise RuntimeError(f"对象没有{main_key_column=}\n原始数据:{row}")
                            return_data.add(line[main_key_column])
                return return_data
            
        case "json":
            with file_name.open("r", encoding=encoding) as f:
                data = json.load(f)
                assert isinstance(data, list), "理论上，这里应该是list[dict]结构，但是不是,请报告 https://github.com/ciaoyizhen/caoyizhen_basetool 让我知道!!!"
                if isinstance(return_data, dict):
                    for row in tqdm(data, disable=disable_tqdm):
                        if main_key_column not in row:
                            raise RuntimeError(f"对象没有{main_key_column=}\n原始数据:{row}")
                        value = row[main_key_column]
                        return_data[value] = row     
                            
                    return return_data
                elif isinstance(return_data, set):
                    for row in tqdm(data, disable=disable_tqdm):
                        if main_key_column not in row:
                            raise RuntimeError(f"对象没有{main_key_column=}\n原始数据:{row}")
                        return_data.add(row[main_key_column])
                    return return_data
                        
                elif isinstance(return_data, list):
                    return data
                
        case "xlsx":
            # 这里导包, 可以让不用pandas时不安装包
            import pandas as pd
            data = pd.read_excel(file_name, **kwargs)
            
            if isinstance(return_data, list):
                for _, row in tqdm(data.iterrows(), total=data.shape[0], disable=disable_tqdm):
                    row = row.to_dict()
                    return_data.append(row)
            elif isinstance(return_data, dict):
                for _, row in tqdm(data.iterrows(), total=data.shape[0], disable=disable_tqdm):
                    row = row.to_dict()
                    if main_key_column not in row:
                        raise RuntimeError(f"对象没有{main_key_column=}\n原始数据:{row}")
                    value = row[main_key_column]
                    return_data[value] = row     
            elif isinstance(return_data, set):
                for _, row in tqdm(data.iterrows(), total=data.shape[0], disable=disable_tqdm):
                    row = row.to_dict()
                    if main_key_column not in row:
                        raise RuntimeError(f"对象没有{main_key_column=}\n原始数据:{row}")
                    return_data.add(row[main_key_column])
            return return_data
            
                
        case "csv":
            import pandas as pd
            if encoding == "utf-8":  # 解决读取csv的编码问题
                data = pd.read_csv(file_name, **kwargs)
            else:
                data = pd.read_csv(file_name, encoding=encoding, **kwargs)
            
            if isinstance(return_data, list):
                for _, row in tqdm(data.iterrows(), total=data.shape[0], disable=disable_tqdm):
                    row = row.to_dict()
                    return_data.append(row)
            elif isinstance(return_data, dict):
                for _, row in tqdm(data.iterrows(), total=data.shape[0], disable=disable_tqdm):
                    row = row.to_dict()
                    
                    if main_key_column not in row:
                        raise RuntimeError(f"对象没有{main_key_column=}\n原始数据:{row}")
                    value = row[main_key_column]
                    return_data[value] = row     
            elif isinstance(return_data, set):
                for _, row in tqdm(data.iterrows(), total=data.shape[0], disable=disable_tqdm):
                    row = row.to_dict()
                    if main_key_column not in row:
                        raise RuntimeError(f"对象没有{main_key_column=}\n原始数据:{row}")
                    return_data.add(row[main_key_column])
                return return_data
            return return_data
            
        case _:
            raise RuntimeError(f"无法识别后缀{file_type=}是什么格式的文件,请传入file_type来控制或修改后缀名")
        

def save_file(file_name: str|Path, data: list, file_type=None, *, encoding="utf-8", ensure_ascii=False, json_indent=4, pd_index=False,**kwargs):
    if isinstance(file_name, str):
        file_name = Path(file_name)
        
    file_name.parent.mkdir(exist_ok=True, parents=True)
    if file_type is None:
        file_type = file_name.suffix.lstrip(".")
        
    match file_type:
        case "jsonl":
            with file_name.open("w", encoding=encoding) as f:
                for item in data:
                    f.write(json.dumps(item, ensure_ascii=ensure_ascii) + "\n")
            base_logger.info(f"文件保存至 {file_name.resolve(strict=True)} ")
        case "json":
            with file_name.open("w", encoding=encoding) as f:
                json.dump(data, f, ensure_ascii=ensure_ascii, indent=json_indent)
            base_logger.info(f"文件保存至 {file_name.resolve(strict=True)} ")
        case "xlsx":
            import pandas as pd
            data = pd.DataFrame(data)
            data.to_excel(file_name, **kwargs, index=pd_index)
            base_logger.info(f"文件保存至 {file_name.resolve(strict=True)} ")
        case "csv":
            import pandas as pd
            data = pd.DataFrame(data)
            data.to_csv(file_name, **kwargs, index=pd_index)
            base_logger.info(f"文件保存至 {file_name.resolve(strict=True)} ")
        case _:
            raise RuntimeError(f"保存文件识别,无法识别{file_type=},该保存成什么格式")
    


# -*- encoding: utf-8 -*-
# @Time    :   2025/10/11 22:16:26
# @File    :   MultiThreadProcess.py
# @Author  :   ciaoyizhen
# @Contact :   yizhen.ciao@gmail.com
# @Function:   多线程的消费者生产者进程处理

import typer
import shutil
from pathlib import Path
from concurrent.futures import as_completed, ThreadPoolExecutor
from ..utils import LoggerManager
from ..file import save_file, read_file
from tqdm import tqdm



class BaseMultiThreading():
    """
    基类, 实现多线程的消费者生产者的处理, 实现边处理边存储
    """
    def __init__(self, max_workers:int, save_temp_dir:str|Path, single_file_size, finally_save_path: str|Path=None, *, file_type:str|Path=None, force_save=False):
        """_summary_

        Args:
            max_workers (int): 并发数
            save_temp_dir (str|Path): 临时存储文件夹
            single_file_size (int): 临时存储时，单个文件的大小
            finally_save_path (str|Path): 最终完整保存的文件
            file_type (str|Path): 文件存储类型
            force_save (bool): 是否无视目录,进行覆盖操作
        """
        self.max_workers = max_workers
        self.single_file_size = single_file_size
        self.save_temp_dir = save_temp_dir
        self.finally_save_path = finally_save_path
        self.file_type = file_type
        
        if not isinstance(self.save_temp_dir, Path):
            self.save_temp_dir = Path(self.save_temp_dir)

        if self.save_temp_dir.exists() and self.save_temp_dir.is_dir():
            if force_save:
                try:
                    self.save_temp_dir.rmdir()  # 只有当 temp_dir 为空时才会成功
                    print("删除成功")
                except OSError as e:
                    force_save_twice = typer.confirm("当前强制保存的目录非空文件,请确实是否删除")
                    if force_save_twice:
                        shutil.rmtree(self.save_temp_dir)
                    else:
                        raise KeyboardInterrupt("请切换目标目录或手动清理数据后，再执行脚本")
            else:
                raise RuntimeError("临时存储目录已经存在,请先确认是否可删除,若不可删除,请更换目录,若可删除,请手动删除")
        self.save_temp_dir.mkdir(parents=True)

        if (finally_save_path is not None) and (not isinstance(finally_save_path, Path)):
            finally_save_path = Path(finally_save_path)
        if finally_save_path is not None and finally_save_path.exists():
            raise RuntimeError("最终存储文件位置已经存在,请先确认是否可删除,若不可删除,请更换存储位置,若可删除,请手动删除")

        self.file_count = 0  # 文件计数
        
        if self.file_type is None:
            # 如果没有传file_type,就根据finally_save_path的后缀判断
            if finally_save_path is None:
                raise RuntimeError(f"没有传入file_type 以及finally_save_path, 无法推断你需要保存的数据样式")
            else:
                self.file_type = finally_save_path.suffix.lstrip(".")
            
        if self.file_type not in {"json", "jsonl", "xlsx", "csv"}:
            raise RuntimeError(f"传入的file_type不符合要求或你的文件后缀不符合要求")

        self.logger = LoggerManager()
            
    def single_data_process(self):
        """
        这个函数实现单个数据怎么处理，输入是一个数据，进行处理，返回一个数据
        需要用户自定义实现
        """
        raise NotImplementedError(f"未实现函数 single_data_process, 该函数需要解决每个数据要怎么")
    
    def _merge_data(self):
        """
        将临时存储的数据合起来，放到最终存储的位置
        """
        if self.finally_save_path is not None:
            total_data = []
            for file in self.save_temp_dir.iterdir():
                data = read_file(file, disable_tqdm=True)
                total_data.extend(data)
                
            save_file(self.finally_save_path, total_data)
            self.logger.info(f"全部数据成功存储于{self.finally_save_path}")
            
    
    def __call__(self, data:list):
        with ThreadPoolExecutor(max_workers=self.max_workers, thread_name_prefix="线程处理数据") as exec, \
            tqdm(total=len(data), desc=f"{self.max_workers}并发处理中") as p_bar:
            try:
                futures_list = []
                for item in data:
                    future = exec.submit(self.single_data_process, item)
                    future.add_done_callback(lambda x: p_bar.update(1))
                    futures_list.append(future)
                
                processed_data = []  # 存储处理好的数据
                for future in as_completed(futures_list):
                    result = future.result()
                    processed_data.append(result)
                    if len(processed_data) == self.single_file_size:
                        temp_file_path = self.save_temp_dir / f"{self.file_count*self.single_file_size}_{(self.file_count+1)*self.single_file_size}.{self.file_type}"
                        save_file(temp_file_path, processed_data)
                        
                        processed_data.clear()
                        self.file_count += 1
                        self.logger.info(f"文件存储成功,已保存至{temp_file_path}")
                        
                if processed_data:
                    # 保存最后的数据
                    temp_file_path = self.save_temp_dir / f"{self.file_count*self.single_file_size}_{(self.file_count+1)*self.single_file_size}.{self.file_type}"
                    save_file(temp_file_path, processed_data)
            except KeyboardInterrupt:
                self.logger.info("用户中断代码,尝试保存当前剩余的数据,请不要进行任何操作")
                exec.shutdown(cancel_futures=True)
                temp_file_path = self.save_temp_dir / f"{self.file_count*self.single_file_size}_{(self.file_count+1)*self.single_file_size}.{self.file_type}"
                save_file(temp_file_path, processed_data)
            except:
                import traceback
                self.logger(traceback.format_exc())
                exec.shutdown(cancel_futures=True)
                
        self._merge_data()


# -*- encoding: utf-8 -*-
# @Time    :   2025/10/12 12:58:27
# @File    :   log_manage.py
# @Author  :   ciaoyizhen
# @Contact :   yizhen.ciao@gmail.com
# @Function:   日志的封装

import os
from loguru import logger
from tqdm import tqdm

class LoggerManager:
    """简洁生产级 Loguru 封装类
    
    用法：
        log = LoggerManager("logs/app.log")
        log.info("启动成功")
        log.error("错误信息")
    """

    def __init__(
        self,
        log_path: str|None = None,
        level: str = "INFO",
        rotation: str = "10 MB",
        retention: str = "7 days",
        compression: str = "zip",
        enqueue: bool = True,
        console: bool = True,
        *,
        file_format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{file}:{function}:{line}</cyan> - <level>{message}</level>",
        console_format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{file}:{function}:{line}</cyan> - <level>{message}</level>"
    ):
        
        logger.remove()  # 清空默认配置

        # 文件日志
        if log_path is not None:
            os.makedirs(os.path.dirname(log_path), exist_ok=True)
            logger.add(
                log_path,
                rotation=rotation,
                retention=retention,
                compression=compression,
                enqueue=enqueue,  # ✅ 支持多进程
                level=level,
                encoding="utf-8",
                format=file_format,
            )

        # 控制台日志（可选）
        if console:
            logger.add(
                sink=lambda msg: tqdm.write(msg, end=""),
                level=level,
                format=console_format,
            )
        self._logger = logger

    # ↓↓↓ 对 loguru 常用方法的封装 ↓↓↓
    def debug(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).critical(msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).exception(msg, *args, **kwargs)
        
base_logger = LoggerManager()
